# [Apr 2019] stolen directly from scipy so I can test getting an array back
#   https://github.com/scipy/scipy/blob/master/scipy/optimize/tests/test_hungarian.py
#   * change imports to local
#   * skip importing and testing `matrix`
#   * add testing of reduced cost for non-T (reduced matrix for T is different)

# Author: Brian M. Clapper, G. Varoquaux, Lars Buitinck
# License: BSD

import numpy as np
from numpy.testing import assert_array_equal
from pytest import raises as assert_raises

from qcelemental.util.scipy_hungarian import linear_sum_assignment


def test_linear_sum_assignment():
    for cost_matrix, expected_cost, expected_reduced_cost_matrix in [
        # Square
        ([[400, 150, 400],
          [400, 450, 600],
          [300, 225, 300]],
         [150, 400, 300],
         [[250,   0, 175],
          [  0,  50, 125],
          [ 75,   0,   0]]
         ),

        # Rectangular variant
        ([[400, 150, 400, 1],
          [400, 450, 600, 2],
          [300, 225, 300, 3]],
         [150, 2, 300],
         [[102,   0, 102,   0],
          [101, 299, 301,   0],
          [  0,  73,   0,   0]]
         ),

        # Square
        ([[10, 10, 8],
          [9, 8, 1],
          [9, 7, 4]],
         [10, 1, 7],
         [[0, 0, 1],
          [5, 4, 0],
          [2, 0, 0]]
         ),

        # Rectangular variant
        ([[10, 10, 8, 11],
          [9, 8, 1, 1],
          [9, 7, 4, 10]],
         [10, 1, 4],
         [[0, 0, 0, 3],
          [6, 5, 0, 0],
          [3, 1, 0, 6]]
         ),

        # n == 2, m == 0 matrix
        ([[], []],
         [],
         [[], []]),
    ]:  # yapf: disable
        cost_matrix = np.array(cost_matrix)
        (row_ind, col_ind), reduced_cost_matrix = linear_sum_assignment(cost_matrix, return_cost=True)
        assert_array_equal(row_ind, np.sort(row_ind))
        assert_array_equal(expected_cost, cost_matrix[row_ind, col_ind])
        assert_array_equal(expected_reduced_cost_matrix, reduced_cost_matrix)

        cost_matrix = cost_matrix.T
        row_ind, col_ind = linear_sum_assignment(cost_matrix)
        assert_array_equal(row_ind, np.sort(row_ind))
        assert_array_equal(np.sort(expected_cost), np.sort(cost_matrix[row_ind, col_ind]))


def test_linear_sum_assignment_input_validation():
    assert_raises(ValueError, linear_sum_assignment, [1, 2, 3])

    C = [[1, 2, 3], [4, 5, 6]]
    assert_array_equal(linear_sum_assignment(C), linear_sum_assignment(np.asarray(C)))
    # assert_array_equal(linear_sum_assignment(C),
    #                    linear_sum_assignment(matrix(C)))

    I = np.identity(3)
    assert_array_equal(linear_sum_assignment(I.astype(np.bool)), linear_sum_assignment(I))
    assert_raises(ValueError, linear_sum_assignment, I.astype(str))

    I[0][0] = np.nan
    assert_raises(ValueError, linear_sum_assignment, I)

    I = np.identity(3)
    I[1][1] = np.inf
    assert_raises(ValueError, linear_sum_assignment, I)



import json
from pathlib import Path
from typing import Any, Dict, Optional, Set, Union

import numpy as np
from pydantic import BaseModel, BaseSettings

from qcelemental.testing import compare_recursive
from qcelemental.util import deserialize, serialize
from qcelemental.util.autodocs import AutoPydanticDocGenerator


def _proto_repr_style(self, join_str: str) -> str:
    """
    This is required to be __repr_str__ as pydantic decided to rebuild a new base
    """
    if self.__config__.repr_style == "default":
        return super().__repr_str__(join_str)
    elif self.__config__.repr_style == 'dict':
        return join_str.join(f'{k}={v}' for k, v in self.dict().items())
    else:
        raise "Invalid repr_style detected, please edit this classes config."


class ProtoModel(BaseModel):
    class Config:
        allow_mutation: bool = False
        extra: str = "forbid"
        json_encoders: Dict[str, Any] = {np.ndarray: lambda v: v.flatten().tolist()}
        serialize_default_excludes: Set = set()
        serialize_skip_defaults: bool = False
        force_skip_defaults: bool = False
        repr_style: str = 'dict'

    def __init_subclass__(cls) -> None:
        cls.__doc__ = AutoPydanticDocGenerator(cls, always_apply=True)
        cls.__repr_str__ = _proto_repr_style

    @classmethod
    def parse_raw(cls, data: Union[bytes, str], *, encoding: str = None) -> 'ProtoModel':  # type: ignore
        """
        Parses raw string or bytes into a Model object.

        Parameters
        ----------
        data : Union[bytes, str]
            A serialized data blob to be deserialized into a Model.
        encoding : str, optional
            The type of the serialized array, available types are: {'json', 'json-ext', 'msgpack-ext', 'pickle'}

        Returns
        -------
        Model
            The requested model from a serialized format.
        """

        if encoding is None:
            if isinstance(data, str):
                encoding = "json"
            elif isinstance(data, bytes):
                encoding = "msgpack-ext"
            else:
                raise TypeError("Input is neither str nor bytes, please specify an encoding.")

        if encoding.endswith(('json', 'javascript', 'pickle')):
            return super().parse_raw(data, content_type=encoding)
        elif encoding in ["msgpack-ext", "json-ext"]:
            obj = deserialize(data, encoding)
        else:
            raise TypeError(f"Content type '{encoding}' not understood.")

        return cls.parse_obj(obj)

    @classmethod
    def parse_file(cls, path: Union[str, Path], *, encoding: str = None) -> 'ProtoModel':  # type: ignore
        """Parses a file into a Model object.

        Parameters
        ----------
        path : Union[str, Path]
            The path to the file.
        encoding : str, optional
            The type of the files, available types are: {'json', 'msgpack', 'pickle'}. Attempts to
            automatically infer the file type from the file extension if None.

        Returns
        -------
        Model
            The requested model from a serialized format.

        """
        path = Path(path)
        if encoding is None:
            if path.suffix in [".json", ".js"]:
                encoding = "json"
            elif path.suffix in [".msgpack"]:
                encoding = "msgpack-ext"
            elif path.suffix in [".pickle"]:
                encoding = "pickle"
            else:
                raise TypeError("Could not infer `encoding`, please provide a `encoding` for this file.")

        return cls.parse_raw(path.read_bytes(), encoding=encoding)

    def dict(self, **kwargs) -> Dict[str, Any]:
        encoding = kwargs.pop("encoding", None)

        kwargs["exclude"] = (
            (kwargs.get("exclude", None) or set()) | self.__config__.serialize_default_excludes)  # type: ignore
        kwargs.setdefault("exclude_unset", self.__config__.serialize_skip_defaults)  # type: ignore
        if self.__config__.force_skip_defaults:  # type: ignore
            kwargs["exclude_unset"] = True

        data = super().dict(**kwargs)

        if encoding is None:
            return data
        elif encoding == "json":
            return json.loads(serialize(data, encoding="json"))
        else:
            raise KeyError(f"Unknown encoding type '{encoding}', valid encoding types: 'json'.")

    def serialize(self,
                  encoding: str,
                  *,
                  include: Optional[Set[str]] = None,
                  exclude: Optional[Set[str]] = None,
                  exclude_unset: bool = False) -> Union[bytes, str]:
        """Generates a serialized representation of the model

        Parameters
        ----------
        encoding : str
            The serialization type, available types are: {'json', 'json-ext', 'msgpack-ext'}
        include : Optional[Set[str]], optional
            Fields to be included in the serialization.
        exclude : Optional[Set[str]], optional
            Fields to be excluded in the serialization.
        exclude_unset : bool, optional
            If True, skips fields that have default values provided.

        Returns
        -------
        Union[bytes, str]
            The serialized model.
        """
        data = self.dict(include=include, exclude=exclude, exclude_unset=exclude_unset)

        return serialize(data, encoding=encoding)

    def compare(self, other: Union['ProtoModel', BaseModel], **kwargs) -> bool:
        """Compares the current object to the provided object recursively.

        Parameters
        ----------
        other : Model
            The model to compare to.
        **kwargs
            Additional kwargs to pass to ``qcelemental.compare_recursive``.

        Returns
        -------
        bool
            True if the objects match.
        """
        return compare_recursive(self, other, **kwargs)


class AutodocBaseSettings(BaseSettings):
    def __init_subclass__(cls) -> None:
        cls.__doc__ = AutoPydanticDocGenerator(cls, always_apply=True)



from enum import Enum
from typing import Any, Dict, List, Set, Optional, Union

import numpy as np
from pydantic import Field, constr, validator

from ..util import provenance_stamp
from .basemodels import ProtoModel
from .basis import BasisSet
from .common_models import ComputeError, DriverEnum, Model, Provenance, qcschema_input_default, qcschema_output_default
from .molecule import Molecule
from .types import Array


class ResultProperties(ProtoModel):
    """
    Named properties of quantum chemistry computations following the MolSSI QCSchema.
    """

    # Calcinfo
    calcinfo_nbasis: Optional[int] = Field(None, description="The number of basis functions for the computation.")
    calcinfo_nmo: Optional[int] = Field(None, description="The number of molecular orbitals for the computation.")
    calcinfo_nalpha: Optional[int] = Field(None, description="The number of alpha electrons in the computation.")
    calcinfo_nbeta: Optional[int] = Field(None, description="The number of beta electrons in the computation.")
    calcinfo_natom: Optional[int] = Field(None, description="The number of atoms in the computation.")

    # Canonical
    nuclear_repulsion_energy: Optional[float] = Field(None, description="The nuclear repulsion energy energy.")
    return_energy: Optional[float] = Field(
        None, description="The energy of the requested method, identical to `return_value` for energy computations.")

    # SCF Keywords
    scf_one_electron_energy: Optional[float] = Field(
        None, description="The one-electron (core Hamiltonian) energy contribution to the total SCF energy.")
    scf_two_electron_energy: Optional[float] = Field(
        None, description="The two-electron energy contribution to the total SCF energy.")
    scf_vv10_energy: Optional[float] = Field(
        None, description="The VV10 functional energy contribution to the total SCF energy.")
    scf_xc_energy: Optional[float] = Field(
        None, description="The functional (XC) energy contribution to the total SCF energy.")
    scf_dispersion_correction_energy: Optional[float] = Field(
        None,
        description="The dispersion correction appended to an underlying functional when a DFT-D method is requested.")
    scf_dipole_moment: Optional[List[float]] = Field(None, description="The X, Y, and Z dipole components.")
    scf_total_energy: Optional[float] = Field(
        None, description="The total electronic energy of the SCF stage of the calculation.")
    scf_iterations: Optional[int] = Field(None, description="The number of SCF iterations taken before convergence.")

    # MP2 Keywords
    mp2_same_spin_correlation_energy: Optional[float] = Field(
        None, description="The portion of MP2 doubles correlation energy from same-spin (i.e. triplet) correlations.")
    mp2_opposite_spin_correlation_energy: Optional[float] = Field(
        None,
        description="The portion of MP2 doubles correlation energy from opposite-spin (i.e. singlet) correlations.")
    mp2_singles_energy: Optional[float] = Field(
        None, description="The singles portion of the MP2 correlation energy. Zero except in ROHF.")
    mp2_doubles_energy: Optional[float] = Field(
        None,
        description=
        "The doubles portion of the MP2 correlation energy including same-spin and opposite-spin correlations.")
    mp2_total_correlation_energy: Optional[float] = Field(
        None, description="The MP2 correlation energy.")  # Old name, to be deprecated
    mp2_correlation_energy: Optional[float] = Field(None, description="The MP2 correlation energy.")
    mp2_total_energy: Optional[float] = Field(None,
                                              description="The total MP2 energy (MP2 correlation energy + HF energy).")
    mp2_dipole_moment: Optional[List[float]] = Field(None, description="The MP2 X, Y, and Z dipole components.")

    # CCSD Keywords
    ccsd_same_spin_correlation_energy: Optional[float] = Field(
        None, description="The portion of CCSD doubles correlation energy from same-spin (i.e. triplet) correlations.")
    ccsd_opposite_spin_correlation_energy: Optional[float] = Field(
        None,
        description="The portion of CCSD doubles correlation energy from opposite-spin (i.e. singlet) correlations")
    ccsd_singles_energy: Optional[float] = Field(
        None, description="The singles portion of the CCSD correlation energy. Zero except in ROHF.")
    ccsd_doubles_energy: Optional[float] = Field(
        None,
        description=
        "The doubles portion of the CCSD correlation energy including same-spin and opposite-spin correlations.")
    ccsd_correlation_energy: Optional[float] = Field(None, description="The CCSD correlation energy.")
    ccsd_total_energy: Optional[float] = Field(
        None, description="The total CCSD energy (CCSD correlation energy + HF energy).")
    ccsd_dipole_moment: Optional[List[float]] = Field(None, description="The CCSD X, Y, and Z dipole components.")
    ccsd_iterations: Optional[int] = Field(None, description="The number of CCSD iterations taken before convergence.")

    # CCSD(T) keywords
    ccsd_prt_pr_correlation_energy: Optional[float] = Field(None, description="The CCSD(T) correlation energy.")
    ccsd_prt_pr_total_energy: Optional[float] = Field(
        None, description="The total CCSD(T) energy (CCSD(T) correlation energy + HF energy).")
    ccsd_prt_pr_dipole_moment: Optional[List[float]] = Field(None,
                                                             description="The CCSD(T) X, Y, and Z dipole components.")

    class Config(ProtoModel.Config):
        force_skip_defaults = True


class WavefunctionProperties(ProtoModel):

    # Class properties
    _return_results_names: Set[str] = {
        'orbitals_a', 'orbitals_b', 'density_a', 'density_b', 'fock_a', 'fock_b', 'eigenvalues_a', 'eigenvalues_b',
        'occupations_a', 'occupations_b'
    }

    # The full basis set description of the quantities
    basis: BasisSet = Field(..., description=str(BasisSet.__doc__))
    restricted: bool = Field(
        ...,
        description=str(
            "If the computation was restricted or not (alpha == beta). If True, all beta quantities are skipped."))

    # Core Hamiltonian
    h_core_a: Optional[Array[float]] = Field(None, description="Alpha-spin core (one-electron) Hamiltonian.")
    h_core_b: Optional[Array[float]] = Field(None, description="Beta-spin core (one-electron) Hamiltonian.")
    h_effective_a: Optional[Array[float]] = Field(None,
                                                  description="Alpha-spin effective core (one-electron) Hamiltonian.")
    h_effective_b: Optional[Array[float]] = Field(None,
                                                  description="Beta-spin effective core (one-electron) Hamiltonian ")

    # SCF Results
    scf_orbitals_a: Optional[Array[float]] = Field(None, description="SCF alpha-spin orbitals.")
    scf_orbitals_b: Optional[Array[float]] = Field(None, description="SCF beta-spin orbitals.")
    scf_density_a: Optional[Array[float]] = Field(None, description="SCF alpha-spin density matrix.")
    scf_density_b: Optional[Array[float]] = Field(None, description="SCF beta-spin density matrix.")
    scf_fock_a: Optional[Array[float]] = Field(None, description="SCF alpha-spin Fock matrix.")
    scf_fock_b: Optional[Array[float]] = Field(None, description="SCF beta-spin Fock matrix.")
    scf_eigenvalues_a: Optional[Array[float]] = Field(None, description="SCF alpha-spin eigenvalues.")
    scf_eigenvalues_b: Optional[Array[float]] = Field(None, description="SCF beta-spin eigenvalues.")
    scf_occupations_a: Optional[Array[float]] = Field(None, description="SCF alpha-spin occupations.")
    scf_occupations_b: Optional[Array[float]] = Field(None, description="SCF beta-spin occupations.")

    # Return results, must be defined last
    orbitals_a: Optional[str] = Field(None, description="Index to the alpha-spin orbitals of the primary return.")
    orbitals_b: Optional[str] = Field(None, description="Index to the beta-spin orbitals of the primary return.")
    density_a: Optional[str] = Field(None, description="Index to the alpha-spin density of the primary return.")
    density_b: Optional[str] = Field(None, description="Index to the beta-spin density of the primary return.")
    fock_a: Optional[str] = Field(None, description="Index to the alpha-spin Fock matrix of the primary return.")
    fock_b: Optional[str] = Field(None, description="Index to the beta-spin Fock matrix of the primary return.")
    eigenvalues_a: Optional[str] = Field(None,
                                         description="Index to the alpha-spin eigenvalues of the primary return.")
    eigenvalues_b: Optional[str] = Field(None, description="Index to the beta-spin eigenvalues of the primary return.")
    occupations_a: Optional[str] = Field(
        None, description="Index to the alpha-spin orbital eigenvalues of the primary return.")
    occupations_b: Optional[str] = Field(
        None, description="Index to the beta-spin orbital eigenvalues of the primary return.")

    class Config(ProtoModel.Config):
        force_skip_defaults = True

    @validator('scf_eigenvalues_a', 'scf_eigenvalues_b', 'scf_occupations_a', 'scf_occupations_b')
    def _assert1d(cls, v, values):

        try:
            v = v.reshape(-1)
        except (ValueError, AttributeError):
            raise ValueError("Vector must be castable to shape (-1, )!")
        return v

    @validator('scf_orbitals_a', 'scf_orbitals_b')
    def _assert2d_nao_x(cls, v, values):
        bas = values.get("basis", None)

        # Do not raise multiple errors
        if bas is None:
            return v

        try:
            v = v.reshape(bas.nbf, -1)
        except (ValueError, AttributeError):
            raise ValueError("Matrix must be castable to shape (nbf, -1)!")
        return v

    @validator(
        'h_core_a',
        'h_core_b',
        'h_effective_a',
        'h_effective_b',

        # SCF
        'scf_density_a',
        'scf_density_b',
        'scf_fock_a',
        'scf_fock_b')
    def _assert2d(cls, v, values):
        bas = values.get("basis", None)

        # Do not raise multiple errors
        if bas is None:
            return v

        try:
            v = v.reshape(bas.nbf, bas.nbf)
        except (ValueError, AttributeError):
            raise ValueError("Matrix must be castable to shape (nbf, nbf)!")
        return v

    @validator('orbitals_a', 'orbitals_b', 'density_a', 'density_b', 'fock_a', 'fock_b', 'eigenvalues_a',
               'eigenvalues_b', 'occupations_a', 'occupations_b')
    def _assert_exists(cls, v, values):

        if values.get(v, None) is None:
            raise ValueError(f"Return quantity {v} does not exist in the values.")
        return v


class WavefunctionProtocolEnum(str, Enum):
    """
    Wavefunction to keep from a Result computation.
    """
    all = "all"
    orbitals_and_eigenvalues = "orbitals_and_eigenvalues"
    return_results = "return_results"
    none = "none"


class ResultProtocols(ProtoModel):
    """
    Protocols regarding the manipulation of a Result output data.
    """

    wavefunction: WavefunctionProtocolEnum = Field(WavefunctionProtocolEnum.none,
                                                   description=str(WavefunctionProtocolEnum.__doc__))

    class Config:
        force_skip_defaults = True


### Primary models


class ResultInput(ProtoModel):
    """The MolSSI Quantum Chemistry Schema"""
    id: Optional[str] = Field(None, description="An optional ID of the ResultInput object.")
    schema_name: constr(strip_whitespace=True, regex=qcschema_input_default) = qcschema_input_default  # type: ignore
    schema_version: int = 1

    molecule: Molecule = Field(..., description="The molecule to use in the computation.")
    driver: DriverEnum = Field(..., description=str(DriverEnum.__doc__))
    model: Model = Field(..., description=str(Model.__base_doc__))
    keywords: Dict[str, Any] = Field({}, description="The program specific keywords to be used.")
    protocols: ResultProtocols = Field(ResultProtocols(), description=str(ResultProtocols.__base_doc__))

    extras: Dict[str, Any] = Field({}, description="Extra fields that are not part of the schema.")

    provenance: Provenance = Field(Provenance(**provenance_stamp(__name__)), description=str(Provenance.__base_doc__))

    def __str__(self):
        return (f"{self.__class__.__name__}"
                f"(driver='{self.driver}' "
                f"model='{self.model.dict()}' "
                f"molecule_hash='{self.molecule.get_hash()[:7]}')")


class Result(ResultInput):
    schema_name: constr(strip_whitespace=True, regex=qcschema_output_default) = qcschema_output_default  # type: ignore

    properties: ResultProperties = Field(..., description=str(ResultProperties.__base_doc__))
    wavefunction: Optional[WavefunctionProperties] = Field(None, description=str(WavefunctionProperties.__base_doc__))

    return_result: Union[float, Array[float], Dict[str, Any]] = Field(
        ..., description="The value requested by the 'driver' attribute.")  # type: ignore

    stdout: Optional[str] = Field(None, description="The standard output of the program.")
    stderr: Optional[str] = Field(None, description="The standard error of the program.")

    success: bool = Field(
        ..., description="The success of a given programs execution. If False, other fields may be blank.")
    error: Optional[ComputeError] = Field(None, description=str(ComputeError.__base_doc__))
    provenance: Provenance = Field(..., description=str(Provenance.__base_doc__))

    @validator("schema_name", pre=True)
    def _input_to_output(cls, v):
        """If qcschema_input is passed in, cast it to output, otherwise no"""
        if v.lower().strip() in [qcschema_input_default, qcschema_output_default]:
            return qcschema_output_default
        raise ValueError("Only {0} or {1} is allowed for schema_name, "
                         "which will be converted to {0}".format(qcschema_output_default, qcschema_input_default))

    @validator("return_result")
    def _validate_return_result(cls, v, values):
        if values["driver"] == "gradient":
            v = np.asarray(v).reshape(-1, 3)
        elif values["driver"] == "hessian":
            v = np.asarray(v)
            nsq = int(v.size**0.5)
            v.shape = (nsq, nsq)

        return v

    @validator('wavefunction', pre=True)
    def _wavefunction_protocol(cls, value, values):

        # We are pre, gotta do extra checks
        if value is None:
            return value
        elif isinstance(value, dict):
            wfn = value.copy()
        elif isinstance(value, WavefunctionProperties):
            wfn = value.dict()
        else:
            raise ValueError('wavefunction must be None, a dict, or a WavefunctionProperties object.')

        # Do not propogate validation errors
        if 'protocols' not in values:
            raise ValueError("Protocols was not properly formed.")

        # Handle restricted
        restricted = wfn.get('restricted', None)
        if restricted is None:
            raise ValueError('`restricted` is required.')

        if restricted:
            for k in list(wfn.keys()):
                if k.endswith('_b'):
                    wfn.pop(k)

        # Handle protocols
        wfnp = values['protocols'].wavefunction
        return_keep = None
        if wfnp == 'all':
            pass
        elif wfnp == 'none':
            wfn = None
        elif wfnp == 'return_results':
            return_keep = [
                'orbitals_a', 'orbitals_b', 'density_a', 'density_b', 'fock_a', 'fock_b', 'eigenvalues_a',
                'eigenvalues_b', 'occupations_a', 'occupations_b'
            ]
        elif wfnp == 'orbitals_and_eigenvalues':
            return_keep = ['orbitals_a', 'orbitals_b', 'eigenvalues_a', 'eigenvalues_b']
        else:
            raise ValueError(f"Protocol `wavefunction:{wfnp}` is not understood.")

        if return_keep is not None:
            ret_wfn = {"restricted": restricted}
            if "basis" in wfn:
                ret_wfn["basis"] = wfn["basis"]

            for rk in return_keep:
                key = wfn.get(rk, None)
                if key is None:
                    continue

                ret_wfn[rk] = key
                ret_wfn[key] = wfn[key]

            return ret_wfn
        else:
            return wfn



import copy

import numpy as np
import pytest

import qcelemental as qcel
from qcelemental.models import basis

center_data = {
    "bs_sto3g_h": {
        "electron_shells": [{
            "harmonic_type": "spherical",
            "angular_momentum": [0],
            "exponents": [3.42525091, 0.62391373, 0.16885540],
            "coefficients": [[0.15432897, 0.53532814, 0.44463454]]
        }]
    },
    "bs_sto3g_o": {
        "electron_shells": [{
            "harmonic_type": "spherical",
            "angular_momentum": [0],
            "exponents": [130.70939, 23.808861, 6.4436089],
            "coefficients": [[0.15432899, 0.53532814, 0.44463454]]
        }, {
            "harmonic_type":
            "cartesian",
            "angular_momentum": [0, 1],
            "exponents": [5.0331513, 1.1695961, 0.3803890],
            "coefficients": [[-0.09996723, 0.39951283, 0.70011547], [0.15591629, 0.60768379, 0.39195739]]
        }, {
            "harmonic_type":
            "cartesian",
            "angular_momentum": [0],
            "exponents": [5.0331513, 1.1695961, 0.3803890],
            "coefficients": [[-5.09996723, 0.39951283, 0.70011547], [0.15591629, 0.60768379, 0.39195739]]
        }]
    },
    "bs_def2tzvp_zr": {
        "electron_shells": [{
            "harmonic_type": "spherical",
            "angular_momentum": [0],
            "exponents": [11.000000000, 9.5000000000, 3.6383667759, 0.76822026698],
            "coefficients": [[-0.19075595259, 0.33895588754, 0.0000000, 0.0000000],
                             [0.0000000, 0.0000000, 1.0000000000, 0.0000000]]
        }, {
            "harmonic_type": "spherical",
            "angular_momentum": [2],
            "exponents": [4.5567957795, 1.2904939799, 0.51646987229],
            "coefficients": [[-0.96190569023E-09, 0.20569990155, 0.41831381851], [0.0000000, 0.0000000, 0.0000000],
                             [0.0000000, 0.0000000, 0.0000000]]
        }, {
            "harmonic_type": "spherical",
            "angular_momentum": [3],
            "exponents": [0.3926100],
            "coefficients": [[1.0000000]]
        }],
        "ecp_electrons": 28,
        "ecp_potentials": [{
            "ecp_type": "scalar",
            "angular_momentum": [0],
            "r_exponents": [2, 2, 2, 2],
            "gaussian_exponents": [7.4880494, 3.7440249, 6.5842120, 3.2921060],
            "coefficients": [[135.15384419, 15.55244130, 19.12219811, 2.43637549]]
        }, {
            "ecp_type": "spinorbit",
            "angular_momentum": [1],
            "r_exponents": [2, 2, 2, 2],
            "gaussian_exponents": [6.4453779, 3.2226886, 6.5842120, 3.2921060],
            "coefficients": [[87.78499169, 11.56406599, 19.12219811, 2.43637549]]
        }]
    }
} # yapf: disable

@pytest.fixture(scope="function")
def result_data_fixture():
    mol = qcel.models.Molecule.from_data("""
        O 0 0 0
        H 0 0 2
        H 0 2 0
    """)

    return {
        "molecule": mol,
        "driver": "energy",
        "model": {
            "method": "UFF"
        },
        "return_result": 5,
        "success": True,
        "properties": {},
        "provenance": {
            "creator": "qcel"
        }
    }


@pytest.fixture(scope="function")
def wavefunction_data_fixture(result_data_fixture):
    bas = basis.BasisSet(name="custom_basis",
                         center_data=center_data,
                         atom_map=["bs_sto3g_o", "bs_sto3g_h", "bs_sto3g_h"])
    c_matrix = np.random.rand(bas.nbf, bas.nbf)
    result_data_fixture["protocols"] = {"wavefunction": "all"}
    result_data_fixture["wavefunction"] = {
        "basis": bas,
        "restricted": True,
        "scf_orbitals_a": c_matrix,
        "orbitals_a": "scf_orbitals_a"
    }

    return result_data_fixture


@pytest.fixture(scope="function")
def optimization_data_fixture(result_data_fixture):

    trajectory = []
    energies = []
    for x in range(5):
        result = result_data_fixture.copy()
        result["return_result"] = x
        trajectory.append(result)
        energies.append(x)

    ret = {
        "initial_molecule": result_data_fixture["molecule"],
        "final_molecule": result_data_fixture["molecule"],
        "trajectory": trajectory,
        "energies": energies,
        "success": True,
        "provenance": {
            "creator": "qcel"
        },
        "input_specification": {
            "model": {
                "method": "UFF"
            }
        }
    }

    return ret


@pytest.mark.parametrize("center_name", center_data.keys())
def test_basis_shell_centers(center_name):
    assert basis.BasisCenter(**center_data[center_name])


def test_basis_set_build():
    bas = basis.BasisSet(name="custom_basis",
                         center_data=center_data,
                         atom_map=["bs_sto3g_o", "bs_sto3g_h", "bs_sto3g_h", "bs_def2tzvp_zr"])

    assert len(bas.center_data) == 3
    assert len(bas.atom_map) == 4
    assert bas.nbf == 21

    es = bas.center_data["bs_sto3g_o"].electron_shells
    assert es[0].is_contracted() is False
    assert es[1].is_contracted() is False
    assert es[2].is_contracted()


def test_basis_electron_center_raises():
    data = center_data["bs_sto3g_h"]["electron_shells"][0].copy()

    # Check bad coefficient length
    bad_coef = data.copy()
    bad_coef["coefficients"] = [[5, 3]]
    with pytest.raises(ValueError) as e:
        basis.ElectronShell(**bad_coef)

    assert "does not match the" in str(e.value)

    # Check bad fused shell
    bad_fused = data.copy()
    bad_fused["angular_momentum"] = [0, 1]
    with pytest.raises(ValueError) as e:
        basis.ElectronShell(**bad_fused)

    assert "fused shell" in str(e.value)


def test_basis_ecp_center_raises():
    # Check coefficients
    data = center_data["bs_def2tzvp_zr"]["ecp_potentials"][0].copy()
    data["coefficients"] = [[5, 3]]

    with pytest.raises(ValueError):
        basis.ECPPotential(**data)

    # Check gaussian_exponents
    data = center_data["bs_def2tzvp_zr"]["ecp_potentials"][0].copy()
    data["gaussian_exponents"] = [5, 3]

    with pytest.raises(ValueError):
        basis.ECPPotential(**data)


def test_basis_map_raises():

    with pytest.raises(ValueError) as e:
        assert basis.BasisSet(name="custom_basis", center_data=center_data, atom_map=["something_odd"])


def test_result_build(result_data_fixture):
    ret = qcel.models.Result(**result_data_fixture)
    assert ret.wavefunction is None


def test_result_build_wavefunction_delete(wavefunction_data_fixture):
    del wavefunction_data_fixture["protocols"]
    ret = qcel.models.Result(**wavefunction_data_fixture)
    assert ret.wavefunction is None


def test_wavefunction_build(wavefunction_data_fixture):
    assert qcel.models.Result(**wavefunction_data_fixture)


def test_wavefunction_matrix_size_error(wavefunction_data_fixture):

    wavefunction_data_fixture["wavefunction"]["scf_orbitals_a"] = np.random.rand(2, 2)
    with pytest.raises(ValueError) as e:
        qcel.models.Result(**wavefunction_data_fixture)

    assert "castable to shape" in str(e.value)


def test_wavefunction_return_result_pointer(wavefunction_data_fixture):

    del wavefunction_data_fixture["wavefunction"]["scf_orbitals_a"]
    with pytest.raises(ValueError) as e:
        qcel.models.Result(**wavefunction_data_fixture)

    assert "does not exist" in str(e.value)


@pytest.mark.parametrize("protocol, restricted, provided, expected", [
    ('none', True, ['orbitals_a', 'orbitals_b'], []),
    (None, True, ['orbitals_a', 'orbitals_b'], []),
    ('all', False, ['orbitals_a', 'orbitals_b'], ['orbitals_a', 'orbitals_b']),
    ('all', True, ['orbitals_a', 'orbitals_b'], ['orbitals_a']),
    ('orbitals_and_eigenvalues', False, ['orbitals_a', 'orbitals_b', 'fock_a', 'fock_b'], ['orbitals_a', 'orbitals_b'
                                                                                           ]),
    ('orbitals_and_eigenvalues', True, ['orbitals_a', 'orbitals_b', 'eigenvalues_a', 'fock_a', 'fock_b'
                                        ], ['orbitals_a', 'eigenvalues_a']),
    ('return_results', True, ['orbitals_a', 'fock_a', 'fock_b'], ['orbitals_a', 'fock_a']),
])
def test_wavefunction_protocols(protocol, restricted, provided, expected, wavefunction_data_fixture):

    wfn_data = wavefunction_data_fixture["wavefunction"]

    if protocol is None:
        wavefunction_data_fixture.pop("protocols")
    else:
        wavefunction_data_fixture["protocols"]["wavefunction"] = protocol

    wfn_data["restricted"] = restricted
    bas = wfn_data["basis"]

    for name in provided:
        scf_name = "scf_" + name
        wfn_data[name] = scf_name
        if "eigen" in name:
            wfn_data[scf_name] = np.random.rand(bas.nbf)
        else:
            wfn_data[scf_name] = np.random.rand(bas.nbf, bas.nbf)

    wfn = qcel.models.Result(**wavefunction_data_fixture)

    if len(expected) == 0:
        assert wfn.wavefunction is None
    else:
        expected_keys = (set(expected) | {"scf_" + x for x in expected} | {"basis", "restricted"})
        assert wfn.wavefunction.dict().keys() == expected_keys


@pytest.mark.parametrize("keep, indices", [
    (None, [0, 1, 2, 3, 4]),
    ('all', [0, 1, 2, 3, 4]),
    ('initial_and_final', [0, 4]),
    ('final', [4]),
    ('none', []),
])
def test_optimization_trajectory_protocol(keep, indices, optimization_data_fixture):

    if keep is not None:
        optimization_data_fixture["protocols"] = {"trajectory": keep}
    opt = qcel.models.Optimization(**optimization_data_fixture)

    assert len(opt.trajectory) == len(indices)
    for result, index in zip(opt.trajectory, indices):
        assert result.return_result == index
